/* 
   31337 recoder;)

   Copyright (C) 2002 Marek "Tilk" Materzok <tilk@lucifer.geos.pl>

   Ideas: Andrzej Lindnal <sznik@osama.geos.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

*/

#ifdef __GNUC__
#ifndef __USE_GNU
#define __USE_GNU
#endif
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#endif

#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <string>
#include <sstream>
#include <fstream>
#include <iostream>
#include <list>
#include <iomanip>
#include <sys/wait.h>
#include <algorithm>
#include <set>
#include "config.h"

#ifdef HAVE_GETTEXT
#include <libintl.h>
#define _(X) gettext(X)
#else
#define _(X) X
#endif

using namespace std;

enum EncodeType { eNumeric, eEscapedHex, eEscapedOctal, eROT, eB64, eRoman };

struct Recoder
{
	Recoder(EncodeType p_et, int param = 0)
	{
		et = p_et;
		prm = param;
	};

	EncodeType et;
	int prm;
};

list<Recoder> recodes;

bool encode = true;
EncodeType etype = eNumeric;
bool f_stdin = true;
bool f_file = false;
int numsys = 2;
int dpc, maxpow;
string ofile;

class Options: private set<string>
{
	public:
		void Parse(const string &str)
		{
			for (string::const_iterator iter = str.begin(); iter != str.end(); )
			{
				string::const_iterator eiter = ::find(iter, str.end(), ',');
				string subs(iter, eiter);
				AddOpt(subs);
				iter = eiter;
				if (iter != str.end()) ++iter;
			}
		}
		void AddOpt(const string &str)
		{
			insert(str);
		}
		void DelOpt(const string &str)
		{
			erase(str);
		}
		bool HasOpt(const string &str)
		{
			return count(str) > 0;
		}
		string OptVal(string str)
		{
			str += "=";
			set<string>::iterator iter = lower_bound(str);
			if (iter != end() && iter->substr(0, str.size()) == str)
				return iter->substr(str.size(), string::npos);
			return string();
		}
};

Options options;

istream *istr = &cin;
ostream *ostr = &cout;

char
ValToChar(int val)
{
	const char *chars = "0123456789abcdefghijklmnopqrstuvwxyz";
	if (val >= 36 || val < 0) return ' ';
	return chars[val];
};

int
CharToVal(char zn)
{
	int vals[] = {
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		0,	1,	2,	3,	4,	5,	6,	7,	8,	9,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	10,	11,	12,	13,	14,	15,	16,	17,	18,	19,	20,	21,	22,	23,	24,
		25,	26,	27,	28,	29,	30,	31,	32,	33,	34,	35,	-1,	-1,	-1,	-1,	-1,
		-1,	10,	11,	12,	13,	14,	15,	16,	17,	18,	19,	20,	21,	22,	23,	24,
		25,	26,	27,	28,	29,	30,	31,	32,	33,	34,	35,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	0};
	return vals[static_cast<char>(zn)];
}

void
GetDpc(int &dpc, int &maxpow)
{
	dpc = 0;
	int n = 1;
	while (n < 255)
	{
		n *= numsys;
		dpc++;
	}
	maxpow = n / numsys;
}

string
EncodeDigit()
{
	unsigned char zn;
	char uzn;
	istr->get(uzn);
	zn = uzn;
	if (!*istr) return string();

	string ret;
	ret.resize(dpc);

	int xpow = maxpow;
	bool nozeroes = false;
	int firstch = 0;
	if (options.HasOpt("nozeroes"))
		nozeroes = true;
	
	for (int n = 0; n < dpc; n++, xpow /= numsys)
	{
		int val = 0;
		while (zn >= xpow)
		{
			zn -= xpow;
			val++;
		}
		if (val == 0 && firstch == 0 && nozeroes) continue;
		if (firstch == 0 && nozeroes)
		{
			firstch = n;
			nozeroes = false;
		}
		ret[n - firstch] = ValToChar(val); 
	}

	ret.resize(dpc - firstch);
	return ret;
}

void
EncodeNumeric()
{
	GetDpc(dpc, maxpow);	

	bool nospaces = false;
	if (options.HasOpt("nospaces") && !options.HasOpt("nozeroes"))
		nospaces = true;
	bool start = true;

	string spacer = options.OptVal("space");
	if (spacer.empty()) spacer = " ";

	while (*istr)
	{
		string str = EncodeDigit();
		if (str.size() == 0) break;
		if (!nospaces && (!start || (options.HasOpt("startspace")))) 
			*ostr << spacer;
		start = false;
		*ostr << str;
	}

	if (options.HasOpt("endspace")) *ostr << spacer;
}

char
DecodeDigit()
{
	int dread;
	unsigned char res = 0;
	string spacer = options.OptVal("space");
	if (spacer.empty()) spacer = " ";
	
	for (dread = 0; dread < dpc; dread++)
	{
		if (!*istr) break;
		char zn;
		
		istr->get(zn);
	
		int cval = CharToVal(zn);
		if (cval == -1 || cval >= numsys || !*istr)
		{
			if (dread == 0)
			{
				dread--;
				continue;
			}
			dread = dpc;
			break;
		}
		res *= numsys;
		res += CharToVal(zn);
	}
	if (dread != dpc) throw 1;
	return static_cast<char>(res);
}

void
DecodeNumeric()
{
	GetDpc(dpc, maxpow);

	while (*istr)
	{
		try
		{
			*ostr << DecodeDigit();
		}
		catch(...)
		{
			break;
		}
	}
}

void
EncodeEscapedHex()
{
	Options bakopt(options);
	options.DelOpt("nozeroes");
	options.AddOpt("startspace");
	options.DelOpt("endspace");
	options.DelOpt("nospaces");
	options.AddOpt("space=\\x");
	numsys = 16;
	EncodeNumeric();
	options = bakopt;
}

void
DecodeEscapedHex()
{
	numsys = 16;
	GetDpc(dpc, maxpow);
	while (*istr)
	{
		char zn = ' ';
		while (*istr && zn != 'x') istr->get(zn);
		if (!*istr) break;
		try
		{
			*ostr << DecodeDigit();
		}
		catch(...)
		{
			break;
		}
	}
}

void
EncodeEscapedOctal()
{
	Options bakopt(options);
	options.DelOpt("nozeroes");
	options.AddOpt("startspace");
	options.DelOpt("endspace");
	options.DelOpt("nospaces");
	options.AddOpt("space=\\0");
	numsys = 8;
	EncodeNumeric();
	options = bakopt;
}

void
DecodeEscapedOctal()
{
	numsys = 8;
	GetDpc(dpc, maxpow);
	while (*istr)
	{
		char zn = ' ';
		while (*istr && zn != '0') istr->get(zn);
		if (!*istr) break;
		try
		{
			*ostr << DecodeDigit();
		}
		catch(...)
		{
			break;
		}
	}
}

void
RecodeROT ()
{
	int &rotate_key = numsys;
  
	if (!encode)
		rotate_key = -rotate_key;

	bool binary = options.HasOpt("binary");
	bool digits = options.HasOpt("digits");
	if (binary) digits = false;

	while (*istr)
	{
		unsigned char zn; char uzn;
		istr->get(uzn);
		zn = uzn;
		if (!*istr) break;
		int znak = zn;

		znak = zn + rotate_key;
		if (binary)
		{
			while (znak < 0) znak += 256;
			while (znak > 255) znak -= 256;
		}
		else
		{
			if (zn >= 'a' && zn <= 'z')
			{
				while (znak < 'a') znak += 26;
				while (znak > 'z') znak -= 26;
			} else
			if (zn >= 'A' && zn <= 'Z')
			{
				while (znak < 'A') znak += 26;
				while (znak > 'Z') znak -= 26;
			} else
			if (zn >= '0' && zn <= '9' && digits)
			{
				while (znak < '0') znak += 10;
				while (znak > '9') znak -= 10;
			} else
				znak -= rotate_key;
		}

		*ostr << static_cast<char>(znak);
	}
}

void
EncodeB64()
{
	char enctable[] = {
		'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 
		'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 
		'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 
		'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3', 
		'4', '5', '6', '7', '8', '9', '+', '/', '='
	};
	int column = 0;
	char obuf[5];
	obuf[4] = 0;
	while (*istr)
	{
		unsigned char ibuf[3];
		int icnt;

		for (icnt = 0; icnt < 3; icnt++)
		{
			char uzn;
			istr->get(uzn);
			ibuf[icnt] = uzn;
			if (!*istr)
			{
				break;
			}
		}
		if (icnt == 0) break;
		for (int n = icnt; n < 3; n++) ibuf[n] = 0;

		obuf[0] = ibuf[0] >> 2;
		obuf[1] = ((ibuf[0] << 4) & 0x30) | (ibuf[1] >> 4);
		if (icnt >= 2) obuf[2] = ((ibuf[1] << 2) & 0x3c) | (ibuf[2] >> 6); 
		else obuf[2] = 64;
		if (icnt == 3) obuf[3] = ibuf[2] & 0x3f; 
		else obuf[3] = 64;

		for (int n = 0; n < 4; n++)
		{
			obuf[n] = enctable[obuf[n]];
		}
		
		if (column == 76)
		{
			column = 0;
			*ostr << "\n";
		}
		*ostr << obuf;
		column += 4;
	}
}

void
DecodeB64()
{
	char dectable[256] = {
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	62,	0,	0,	0,	63,	
		52,	53,	54,	55,	56,	57,	58,	59,	60,	61,	0,	0,	0,	0,	0,	0,	
		0,	 0,	 1,	 2,	 3,	 4,	 5,	 6,	 7,	 8,	 9,	10,	11,	12,	13,	14,	
		15,	16,	17,	18,	19,	20,	21,	22,	23,	24,	25,	0,	0,	0,	0,	0,	
		0,	26,	27,	28,	29,	30,	31,	32,	33,	34,	35,	36,	37,	38,	39,	40,	
		41,	42,	43,	44,	45,	46,	47,	48,	49,	50,	51,	0,	0,	0,	0,	0,
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	
	};
	
	bool complete = true;
	while (*istr)
	{
		unsigned char ibuf[4];
		for (int n = 0; n < 4; n++)
		{
			char uzn;
			istr->get(uzn);
			ibuf[n] = uzn;
			if (!*istr)
			{
				complete = false;
				break;
			}
			if (ibuf[n] == ' ' || ibuf[n] == '\n' || ibuf[n] == '\t'
					|| ibuf[n] == '\r')
			{
				n--; continue;
			}
		}
		if (!complete) break;
		
		*ostr << static_cast<char>(
			((dectable[ibuf[0]] << 2) | (dectable[ibuf[1]] >> 4)));
		if (ibuf[2] != '=') *ostr << static_cast<char>(
				((dectable[ibuf[1]] << 4) | (dectable[ibuf[2]] >> 2)));
		if (ibuf[3] != '=') *ostr << static_cast<char>(
				((dectable[ibuf[2]] << 6) | (dectable[ibuf[3]])));
	};
}

string
EncodeRomanDigit()
{
	string ret;
	unsigned char zn;
	char uzn;
	istr->get(uzn);
	zn = uzn;
	if (!*istr) return ret;
	if (zn == 0)
	{
		ret = "-";
		return ret;
	}

	char M[4] = {'I', 'X', 'C', 'M'};
	char D[4] = {'V', 'L', 'D', '-'};

	int coef, deg;

	for (coef = 100, deg = 3; deg > 0; deg--, coef /= 10)
	{
		while (zn > 10 * coef)
		{
			ret.push_back(M[deg]);
			zn -= 10 * coef;
		}

		if (zn >= 9 * coef)
		{
			ret.push_back(M[deg - 1]);
			ret.push_back(M[deg]);
			zn -= 9 * coef;
		}

		if (zn >= 5 * coef)
		{
			ret.push_back(D[deg - 1]);
			zn -= 5 * coef;
		}

		if (zn >= 4 * coef)
		{
			ret.push_back(M[deg - 1]);
			ret.push_back(D[deg - 1]);
			zn -= 4 * coef; 
		}
	}
	while (zn--) ret.push_back('I');

	return ret;
}

void
EncodeRoman()
{
	while(*istr)
	{
		string str = EncodeRomanDigit();
		if (!str.empty()) *ostr << str << " ";
	}
}

char
DecodeRomanDigit()
{
	unsigned char val = 0;
	int lastd = 0;
	bool complete = false;
	bool loop = false;

	while (!complete)
	{
		int d = 0;
		char ch;
		istr->get(ch);
		if (!(*istr || loop)) throw 1;
		if (!*istr || isspace(ch))
		{
			ch = ' ';
			complete = true;
		}
		loop = true;
		ch = toupper(ch);
		switch(ch)
		{
			case 'I': d = 1; break;
			case 'V': d = 5; break;
			case 'X': d = 10; break;
			case 'L': d = 50; break;
			case 'C': d = 100; break;
			case 'D': d = 500; break;
			case 'M': d = 1000; break;
			case '-': d = 0; break;
			default: loop = false;
		}
		if (lastd < d) val -= lastd;
		else val += lastd;
		lastd = d;
	}
	
	return static_cast<char>(val);
}

void
DecodeRoman()
{
	while (*istr)
	{
		try
		{
			*ostr << DecodeRomanDigit();
		}
		catch(...)
		{
			return;
		}
	}
}

void
WorkRecode(Recoder &rc)
{
	etype = rc.et;
	numsys = rc.prm;
	switch(etype)
	{
		case eNumeric:
			if (encode) EncodeNumeric(); else DecodeNumeric();
			break;
		case eEscapedHex:
			if (encode) EncodeEscapedHex(); else DecodeEscapedHex();
			break;
		case eEscapedOctal:
			if (encode) EncodeEscapedOctal(); else DecodeEscapedOctal();
			break;
		case eB64:
			if (encode) EncodeB64(); else DecodeB64();
			break;
		case eRoman:
			if (encode) EncodeRoman(); else DecodeRoman();
			break;
		case eROT:
			RecodeROT();
			break;
	};
}

void
DoRecode()
{
	if (recodes.size() == 1)
	{
		WorkRecode(recodes.front());
		return;
	}

	if (!encode) reverse(recodes.begin(), recodes.end());
	
	int pipes[2];

	istream *bak_istr = istr;
	ostream *bak_ostr = ostr;
	istr = &cin;
	ostr = &cout;
	
	pipe(pipes);
	if (fork() == 0)
	{
		istr = bak_istr;
		close(1);
		dup2(pipes[1], 1);
		close(pipes[1]);
		WorkRecode(recodes.front());
		ostr->flush();
		close(1);
		exit(0);
	}

	int readfd = pipes[0];

	for (list<Recoder>::iterator iter = ++recodes.begin();
			iter != --recodes.end(); ++iter)
	{
		close(pipes[1]);
		pipe(pipes);
		if (fork() == 0)
		{
			close(0);
			close(1);
			dup2(readfd, 0);
			dup2(pipes[1], 1);
			close(readfd);
			close(pipes[1]);
			WorkRecode(*iter);
			ostr->flush();
			close(1);
			exit(0);
		}
		close(readfd);
		readfd = pipes[0];
	}

	close(pipes[1]);

	if (fork() == 0)
	{
		ostr = bak_ostr;
		close(0);
		dup2(readfd, 0);
		WorkRecode(recodes.back());
		exit(0);
	}

	close(pipes[0]);

	for (int n = 0; n < recodes.size(); n++) wait(0);
}

int
main(int argc, char **argv)
{
#ifdef __GNUC__
	static struct option long_options[] =
	{
		{"encode", false, 0, 'e'},
		{"decode", false, 0, 'd'},
		{"recode-arg", false, 0, 'a'},
		{"recode-file", false, 0, 'f'},
		{"output-file", true, 0, 'o'},
		{"numeric", true, 0, 'n'},
		{"rot", true, 0, 'r'},
		{"escaped-hex", false, 0, 'H'},
		{"escaped-octal", false, 0, 'O'},
		{"help", false, 0, 'h'},
		{"base64", false, 0, 'B'},
		{"version", false, 0, 'v'},
		{"roman", false, 0, 'R'},
		{"options", true, 0, 'x'},
		0
	};
#endif
#ifdef HAVE_GETTEXT
	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
#endif

	static const char *opt_list = "edEDn:hHOr:Bafo:vRx:";
	int c;
#ifdef __GNUC__
	while ((c = getopt_long(argc, argv, opt_list, long_options, 0)) != -1)
#else
	while ((c = getopt(argc, argv, opt_list)) != -1)
#endif
	{
		switch(c)
		{
			case 'v':
				cerr << _("31337 recoder ;) version ") << VERSION "\n";
				cerr << "Copyleft 2002 Marek \"Tilk\" Materzok <tilk@lucifer.geos.pl>\n";
				cerr << _("Ideas: Andrzej Lindnal <sznik@osama.geos.pl>\n");
				cerr << _("Licensed under GNU GPL version 2\n");
				return 0;
			case 'h':
				cerr << "31337 recoder ;)\n";
				cerr << _("usage:\n"); 
				cerr << "\t" << argv[0] << _(" -e [args]\t= encode from stdin\n");
				cerr << "\t" << argv[0] << _(" -ea [args] text\t= encode arguments\n");
				cerr << "\t" << argv[0] << _(" -ef [args] files...\t= encode from files\n");
				cerr << "\t" << argv[0] << _(" -d [args]\t= decode from stdin\n");
				cerr << "\t" << argv[0] << _(" -da [args] text\t= decode arguments\n");
				cerr << "\t" << argv[0] << _(" -df [args] files...\t= decode from files\n");
				cerr << _("\narguments:\n");
				cerr << _("\t-r num - recode with ROT-num (Caesar shift)\n");
				cerr << _("\t-n num - recode to numeric system num (2-36)\n");
				cerr << _("\t-B - recode with Base64\n");
				cerr << _("\t-H - recode to escaped hexadecimal\n");
				cerr << _("\t-O - recode to escaped octal\n");
				cerr << _("\t-R - recode to Roman numbers\n");
				cerr << _("\t-o file - output to file\n");
				cerr << _("\t-x options - pass options to codecs (look to manpage)\n");
				cerr << "\n";
				return 0;
			case 'e':
				encode = true;
				f_stdin = true;
				break;
			case 'd':
				encode = false;
				f_stdin = true;
				break;
			case 'a':
				f_stdin = false;
				break;
			case 'f':
				f_file = true;
				break;
			case 'n':
				etype = eNumeric;
				numsys = atoi(optarg);
				if (numsys <= 1 || numsys > 36)
				{
					cerr << _("Supported numeric systems: 2 - 36.\n");
					return 1;
				}
				recodes.push_back(Recoder(etype, numsys));
				break;
			case 'r':
				etype = eROT;
				numsys = atoi(optarg);
				if (etype == eROT && (numsys < 1 || numsys > 26))
				{
					cerr << _("Supported ROTs: 1 - 26.\n");
					return 1;
				}
				recodes.push_back(Recoder(etype, numsys));
				break;
			case 'o':
				ofile = optarg;
				break;
			case 'H':
				etype = eEscapedHex;
				recodes.push_back(Recoder(etype));
				break;
			case 'O':
				etype = eEscapedOctal;
				recodes.push_back(Recoder(etype));
				break;
			case 'B':
				etype = eB64;
				recodes.push_back(Recoder(etype));
				break;
			case 'R':
				etype = eRoman;
				recodes.push_back(Recoder(etype));
				break;
			case 'x':
				options.Parse(optarg);
				break;
			case '?':
				return 1;
		}
	}
	
	if (recodes.empty())
	{
		cerr << _("No codecs specified. Use ") << argv[0] << 
			_(" -h for help.\n");
		return 1;
	}

	if (ofile.size() > 0)
	{
		ostr = new ofstream(ofile.c_str());
		if (!*ostr)
		{
			cerr << _("Can't open file ") << ofile << _(" for writing!\n");
			return 1;
		}
	}
	
	if (!f_stdin && !f_file)
	{
		stringstream *sstr = new stringstream();
		for (int i = optind; i < argc; i++)
		{
			*sstr << argv[i];
			if (i != argc - 1) *sstr << " ";
		}
		istr = sstr;
		DoRecode();
		delete sstr;
	}
	else
	if (f_file)
	{
		for (int i = optind; i < argc; i++)
		{
			ifstream *ifstr = new ifstream(argv[i]);
			istr = ifstr;
			if (!*istr)
			{
				cerr << _("Can't open file ") << argv[i] << "!\n";
				continue;
			}
			DoRecode();
			delete ifstr;
		}
	}
	else
	{
		DoRecode();
	}
	ostr->flush();
	cerr << "\n";

	return 0;
}

